import hashlib
import re
from enum import Enum, auto

import requests
from django import forms
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.hashers import PBKDF2PasswordHasher as ph
from django.core.exceptions import ValidationError


def is_password_pwned(password):
    # Create SHA1 hash of the password
    sha1_hash = hashlib.sha1(password.encode()).hexdigest()
    prefix = sha1_hash[:5]  # Take the first 5 characters of the hash as the prefix

    # Make a request to the Pwned Passwords API
    url = f"https://api.pwnedpasswords.com/range/{prefix}"
    response = requests.get(url)

    if response.status_code != 200:
        return False  # backend ignore this and frontend tells the user
        # that we cannot verify this at the moment

    # Check if the password's hash is found in the response body
    hash_suffixes = response.text.split("\r\n")
    for suffix in hash_suffixes:
        # Compare the suffix from the hash to the api response
        # api response is using the format of suffix:count hence
        # we need to get rid of the ending count number
        if sha1_hash[5:].upper() == suffix[:35].upper():
            return True
    return False


class PasswordStrength(Enum):
    STUDENT = auto()
    INDEPENDENT = auto()
    TEACHER = auto()

    def password_test(self, password):
        if password:
            if self is PasswordStrength.STUDENT:
                minimum_password_length = 6
                # Make student password case insensitive
                password = password.lower()
                if not password_strength_test(
                    password=password,
                    minimum_password_length=minimum_password_length,
                    upper=False,
                    lower=False,
                    numbers=False,
                    special_char=False,
                ):
                    raise forms.ValidationError(
                        f"Password not strong enough, consider using at least {minimum_password_length} characters and making it hard to guess."
                    )
                if is_password_pwned(password):
                    raise forms.ValidationError("Password is too common, consider using a different password.")
            elif self is PasswordStrength.INDEPENDENT:
                minimum_password_length = 8
                if not password_strength_test(
                    password=password,
                    minimum_password_length=minimum_password_length,
                    upper=True,
                    lower=True,
                    numbers=True,
                    special_char=False,
                ):
                    raise forms.ValidationError(
                        f"Password not strong enough, consider using at least {minimum_password_length} characters, "
                        "upper and lower case letters, and numbers and making it hard to guess."
                    )
                if is_password_pwned(password):
                    raise forms.ValidationError("Password is too common, consider using a different password.")
            else:
                minimum_password_length = 10
                if not password_strength_test(
                    password=password,
                    minimum_password_length=minimum_password_length,
                    upper=True,
                    lower=True,
                    numbers=True,
                    special_char=True,
                ):
                    raise forms.ValidationError(
                        f"Password not strong enough, consider using at least {minimum_password_length} characters, "
                        "upper and lower case letters, numbers, special characters and making it hard to guess."
                    )
                if is_password_pwned(password):
                    raise forms.ValidationError("Password is too common, consider using a different password.")

        return password


def password_strength_test(
    password,
    minimum_password_length,
    upper=True,
    lower=True,
    numbers=True,
    special_char=True,
):
    most_used_passwords = [
        "Abcd1234",
        "Password1",
        "Qwerty123",
        "password",
        "qwerty",
        "abcdef",
    ]
    return (
        len(password) >= minimum_password_length
        and (not upper or re.search(r"[A-Z]", password))
        and (not lower or re.search(r"[a-z]", password))
        and (not numbers or re.search(r"[0-9]", password))
        and (not special_char or re.search(r"[!@#$%^&*()_+\-=\[\]{};':\"\\|,.<>\/?]", password))
        and (password not in most_used_passwords)
    )


def form_clean_password(self, password_field_name, strength: PasswordStrength):
    password = self.cleaned_data.get(password_field_name, None)
    password = strength.password_test(password)
    if hasattr(self, "user"):
        current_user_password = self.user.password
        algorithm, iterations, salt, hash = current_user_password.split("$")
        new_hashed_password = ph().encode(password, salt)
        if new_hashed_password == current_user_password:
            raise ValidationError(f"Please choose a password that you haven't used before")
    return password


def check_update_password(form, user, request, data):
    changing_password = False
    if data["password"] != "":
        changing_password = True
        user.set_password(data["password"])
        user.save()
        update_session_auth_hash(request, form.user)

    return changing_password



from django import template

register = template.Library()


@register.inclusion_tag("portal/partials/hero_card.html", takes_context=True)
def hero_card(context, hero_card_name):
    """
    Registers the inclusion tag for the hero card partial.
    The template currently expects the following context elements:
    - image: the path to the image at the top of the hero card
    - title: the heading of the hero card
    - description: the text paragraph of the hero card
    - button1: dictionary which contains:
        - text: the text on the card's first button
        - url: the url the first button redirects to
        - url_args: the args needed for the first button's url
    - button2: dictionary which contains:
        - text: the text on the card's second button
        - url: the url the second button redirects to
        - url_args: the args needed for the second button's url
    """
    return context[hero_card_name]



from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from two_factor.views.profile import DisableView

from .form import DisableForm


# This is not changed but imports the from form.py so it overwrites the disable 2FA form
@method_decorator(never_cache, name='dispatch')
class CustomDisableView(DisableView):
    form_class = DisableForm



import re

from django import forms
from django.contrib.auth.forms import (
    PasswordChangeForm,
    UserCreationForm,
    AdminPasswordChangeForm,
)

ADMIN_PASSWORD_TOO_WEAK_MESSAGE = """
Password is too weak. Please choose a password that's at least 14 characters long,
contains at least one lowercase letter, one uppercase letter, one digit and
one special character.
"""

ADMIN_PASSWORD_PATTERN = re.compile(
    "^(?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?=.*[!@#$%^&*()_+\-=\[\]{};':\"\\|,.<>\/?]).{14,}$"
)


class AdminChangeOwnPasswordForm(PasswordChangeForm):
    error_messages = {
        **PasswordChangeForm.error_messages,
        "password_too_weak": ADMIN_PASSWORD_TOO_WEAK_MESSAGE,
    }

    def clean_new_password1(self):
        new_password1 = self.cleaned_data["new_password1"]

        if not re.match(ADMIN_PASSWORD_PATTERN, new_password1):
            raise forms.ValidationError(
                self.error_messages["password_too_weak"],
                code="password_too_weak",
            )

        return new_password1


class AdminUserCreationForm(UserCreationForm):
    error_messages = {
        **UserCreationForm.error_messages,
        "password_too_weak": ADMIN_PASSWORD_TOO_WEAK_MESSAGE,
    }

    def clean_password1(self):
        password1 = self.cleaned_data["password1"]

        if not re.match(ADMIN_PASSWORD_PATTERN, password1):
            raise forms.ValidationError(
                self.error_messages["password_too_weak"],
                code="password_too_weak",
            )

        return password1


class AdminChangeUserPasswordForm(AdminPasswordChangeForm):
    error_messages = {
        **AdminPasswordChangeForm.error_messages,
        "password_too_weak": ADMIN_PASSWORD_TOO_WEAK_MESSAGE,
    }

    def clean_password1(self):
        password1 = self.cleaned_data["password1"]

        if not re.match(ADMIN_PASSWORD_PATTERN, password1):
            raise forms.ValidationError(
                self.error_messages["password_too_weak"],
                code="password_too_weak",
            )

        return password1



from django.template import Context, Template


def test_banner(snapshot):
    test_banner = {
        "title": "Test title",
        "subtitle": "Test subtitle",
        "text": "Test text",
        "image_class": "test--image--class",
        "banner_class": "test--banner--class",
    }

    context = Context({"BANNER": test_banner})
    template_to_render = Template(
        "{% load banner_tags %}" '{% banner banner_name="BANNER" %}'
    )
    rendered_template = template_to_render.render(context)

    snapshot.assert_match(rendered_template)


def test_headline(snapshot):
    test_headline = {"title": "Test title", "description": "Test description"}

    context = Context({"HEADLINE": test_headline})
    template_to_render = Template(
        "{% load headline_tags %}" '{% headline headline_name="HEADLINE" %}'
    )
    rendered_template = template_to_render.render(context)

    snapshot.assert_match(rendered_template)


def test_benefits(snapshot):
    test_benefits = {
        "first": {
            "image": "",
            "title": "Test title",
            "text": "Test text",
            "button": {"text": "Test button", "link": "home"},
        },
        "second": {
            "image": "",
            "title": "Test title",
            "text": "Test text",
            "button": {"text": "Test button", "link": "home"},
        },
        "third": {
            "image": "",
            "title": "Test title",
            "text": "Test text",
            "button": {"text": "Test button", "link": "home"},
        },
    }

    context = Context({"BENEFITS": test_benefits})

    template_to_render = Template("{% load benefits_tags %}" "{% benefits %}")

    rendered_template = template_to_render.render(context)

    snapshot.assert_match(rendered_template)
